import pytest

from jsonschema import ValidationError

from flask_restx import errors, schemas


class SchemasTest:
    def test_lazyness(self):
        schema = schemas.LazySchema("oas-2.0.json")
        assert schema._schema is None

        "" in schema  # Trigger load
        assert schema._schema is not None
        assert isinstance(schema._schema, dict)

    def test_oas2_schema_is_present(self):
        assert hasattr(schemas, "OAS_20")
        assert isinstance(schemas.OAS_20, schemas.LazySchema)


class ValidationTest:
    def test_oas_20_valid(self):
        assert schemas.validate(
            {
                "swagger": "2.0",
                "info": {
                    "title": "An empty minimal specification",
                    "version": "1.0",
                },
                "paths": {},
            }
        )

    def test_oas_20_invalid(self):
        with pytest.raises(schemas.SchemaValidationError) as excinfo:
            schemas.validate(
                {
                    "swagger": "2.0",
                    "should": "not be here",
                }
            )
        for error in excinfo.value.errors:
            assert isinstance(error, ValidationError)

    def test_unknown_schema(self):
        with pytest.raises(errors.SpecsError):
            schemas.validate({"valid": "no"})

    def test_unknown_version(self):
        with pytest.raises(errors.SpecsError):
            schemas.validate({"swagger": "42.0"})



from flask_restx.swagger import extract_path, extract_path_params, parse_docstring


class ExtractPathTest(object):
    def test_extract_static_path(self):
        path = "/test"
        assert extract_path(path) == "/test"

    def test_extract_path_with_a_single_simple_parameter(self):
        path = "/test/<parameter>"
        assert extract_path(path) == "/test/{parameter}"

    def test_extract_path_with_a_single_typed_parameter(self):
        path = "/test/<string:parameter>"
        assert extract_path(path) == "/test/{parameter}"

    def test_extract_path_with_a_single_typed_parameter_with_arguments(self):
        path = "/test/<string(length=2):parameter>"
        assert extract_path(path) == "/test/{parameter}"

    def test_extract_path_with_multiple_parameters(self):
        path = "/test/<parameter>/<string:other>/"
        assert extract_path(path) == "/test/{parameter}/{other}/"


class ExtractPathParamsTestCase(object):
    def test_extract_static_path(self):
        path = "/test"
        assert extract_path_params(path) == {}

    def test_extract_single_simple_parameter(self):
        path = "/test/<parameter>"
        assert extract_path_params(path) == {
            "parameter": {
                "name": "parameter",
                "type": "string",
                "in": "path",
                "required": True,
            }
        }

    def test_single_int_parameter(self):
        path = "/test/<int:parameter>"
        assert extract_path_params(path) == {
            "parameter": {
                "name": "parameter",
                "type": "integer",
                "in": "path",
                "required": True,
            }
        }

    def test_single_float_parameter(self):
        path = "/test/<float:parameter>"
        assert extract_path_params(path) == {
            "parameter": {
                "name": "parameter",
                "type": "number",
                "in": "path",
                "required": True,
            }
        }

    def test_extract_path_with_multiple_parameters(self):
        path = "/test/<parameter>/<int:other>/"
        assert extract_path_params(path) == {
            "parameter": {
                "name": "parameter",
                "type": "string",
                "in": "path",
                "required": True,
            },
            "other": {
                "name": "other",
                "type": "integer",
                "in": "path",
                "required": True,
            },
        }

    def test_extract_parameter_with_arguments(self):
        path = "/test/<string(length=2):parameter>"
        assert extract_path_params(path) == {
            "parameter": {
                "name": "parameter",
                "type": "string",
                "in": "path",
                "required": True,
            }
        }

    # def test_extract_registered_converters(self):
    #     class ListConverter(BaseConverter):
    #         def to_python(self, value):
    #             return value.split(',')

    #         def to_url(self, values):
    #             return ','.join(super(ListConverter, self).to_url(value) for value in values)

    #     self.app.url_map.converters['list'] = ListConverter

    #     path = '/test/<list:parameters>'
    #     with self.context():
    #         self.assertEqual(extract_path_params(path), [{
    #             'name': 'parameters',
    #             'type': 'number',
    #             'in': 'path',
    #             'required': True
    #         }])


class ParseDocstringTest(object):
    def test_empty(self):
        def without_doc():
            pass

        parsed = parse_docstring(without_doc)

        assert parsed["raw"] is None
        assert parsed["summary"] is None
        assert parsed["details"] is None
        assert parsed["returns"] is None
        assert parsed["raises"] == {}
        assert parsed["params"] == []

    def test_single_line(self):
        def func():
            """Some summary"""
            pass

        parsed = parse_docstring(func)

        assert parsed["raw"] == "Some summary"
        assert parsed["summary"] == "Some summary"
        assert parsed["details"] is None
        assert parsed["returns"] is None
        assert parsed["raises"] == {}
        assert parsed["params"] == []

    def test_multi_line(self):
        def func():
            """
            Some summary
            Some details
            """
            pass

        parsed = parse_docstring(func)

        assert parsed["raw"] == "Some summary\nSome details"
        assert parsed["summary"] == "Some summary"
        assert parsed["details"] == "Some details"
        assert parsed["returns"] is None
        assert parsed["raises"] == {}
        assert parsed["params"] == []

    def test_multi_line_and_dot(self):
        def func():
            """
            Some summary. bla bla
            Some details
            """
            pass

        parsed = parse_docstring(func)

        assert parsed["raw"] == "Some summary. bla bla\nSome details"
        assert parsed["summary"] == "Some summary"
        assert parsed["details"] == "bla bla\nSome details"
        assert parsed["returns"] is None
        assert parsed["raises"] == {}
        assert parsed["params"] == []

    def test_raises(self):
        def func():
            """
            Some summary.
            :raises SomeException: in case of something
            """
            pass

        parsed = parse_docstring(func)

        assert (
            parsed["raw"]
            == "Some summary.\n:raises SomeException: in case of something"
        )
        assert parsed["summary"] == "Some summary"
        assert parsed["details"] is None
        assert parsed["returns"] is None
        assert parsed["params"] == []
        assert parsed["raises"] == {"SomeException": "in case of something"}



from collections import OrderedDict
from datetime import date, datetime
from decimal import Decimal
from functools import partial

import pytz
import pytest

from flask import Blueprint
from flask_restx import fields, Api


class FieldTestCase(object):
    field_class = None

    @pytest.fixture
    def api(self, app):
        blueprint = Blueprint("api", __name__)
        api = Api(blueprint)
        app.register_blueprint(blueprint)
        yield api

    def assert_field(self, field, value, expected):
        assert field.output("foo", {"foo": value}) == expected

    def assert_field_raises(self, field, value):
        with pytest.raises(fields.MarshallingError):
            field.output("foo", {"foo": value})


class BaseFieldTestMixin(object):
    def test_description(self):
        field = self.field_class(description="A description")
        assert "description" in field.__schema__
        assert field.__schema__["description"] == "A description"

    def test_title(self):
        field = self.field_class(title="A title")
        assert "title" in field.__schema__
        assert field.__schema__["title"] == "A title"

    def test_required(self):
        field = self.field_class(required=True)
        assert field.required

    def test_readonly(self):
        field = self.field_class(readonly=True)
        assert "readOnly" in field.__schema__
        assert field.__schema__["readOnly"]


class NumberTestMixin(object):
    def test_min(self):
        field = self.field_class(min=0)
        assert "minimum" in field.__schema__
        assert field.__schema__["minimum"] == 0
        assert "exclusiveMinimum" not in field.__schema__

    def test_min_as_callable(self):
        field = self.field_class(min=lambda: 0)
        assert "minimum" in field.__schema__
        assert field.__schema__["minimum"] == 0

    def test_min_exlusive(self):
        field = self.field_class(min=0, exclusiveMin=True)
        assert "minimum" in field.__schema__
        assert field.__schema__["minimum"] == 0
        assert "exclusiveMinimum" in field.__schema__
        assert field.__schema__["exclusiveMinimum"] is True

    def test_max(self):
        field = self.field_class(max=42)
        assert "maximum" in field.__schema__
        assert field.__schema__["maximum"] == 42
        assert "exclusiveMaximum" not in field.__schema__

    def test_max_as_callable(self):
        field = self.field_class(max=lambda: 42)
        assert "maximum" in field.__schema__
        assert field.__schema__["maximum"] == 42

    def test_max_exclusive(self):
        field = self.field_class(max=42, exclusiveMax=True)
        assert "maximum" in field.__schema__
        assert field.__schema__["maximum"] == 42
        assert "exclusiveMaximum" in field.__schema__
        assert field.__schema__["exclusiveMaximum"] is True

    def test_mulitple_of(self):
        field = self.field_class(multiple=5)
        assert "multipleOf" in field.__schema__
        assert field.__schema__["multipleOf"] == 5


class StringTestMixin(object):
    def test_min_length(self):
        field = self.field_class(min_length=1)
        assert "minLength" in field.__schema__
        assert field.__schema__["minLength"] == 1

    def test_min_length_as_callable(self):
        field = self.field_class(min_length=lambda: 1)
        assert "minLength" in field.__schema__
        assert field.__schema__["minLength"] == 1

    def test_max_length(self):
        field = self.field_class(max_length=42)
        assert "maxLength" in field.__schema__
        assert field.__schema__["maxLength"] == 42

    def test_max_length_as_callable(self):
        field = self.field_class(max_length=lambda: 42)
        assert "maxLength" in field.__schema__
        assert field.__schema__["maxLength"] == 42

    def test_pattern(self):
        field = self.field_class(pattern="[a-z]")
        assert "pattern" in field.__schema__
        assert field.__schema__["pattern"] == "[a-z]"


class RawFieldTest(BaseFieldTestMixin, FieldTestCase):
    """Test Raw field AND some common behaviors"""

    field_class = fields.Raw

    def test_type(self):
        field = fields.Raw()
        assert field.__schema__["type"] == "object"

    def test_default(self):
        field = fields.Raw(default="aaa")
        assert field.__schema__["default"] == "aaa"
        self.assert_field(field, None, "aaa")

    def test_default_as_callable(self):
        field = fields.Raw(default=lambda: "aaa")
        assert field.__schema__["default"] == "aaa"
        self.assert_field(field, None, "aaa")

    def test_with_attribute(self):
        field = fields.Raw(attribute="bar")
        assert field.output("foo", {"bar": 42}) == 42

    def test_with_lambda_attribute(self, mocker):
        obj = mocker.Mock()
        obj.value = 42
        field = fields.Raw(attribute=lambda x: x.value)
        assert field.output("foo", obj) == 42

    def test_with_partial_attribute(self, mocker):
        def f(x, suffix):
            return "{0}-{1}".format(x.value, suffix)

        obj = mocker.Mock()
        obj.value = 42

        p = partial(f, suffix="whatever")
        field = fields.Raw(attribute=p)
        assert field.output("foo", obj) == "42-whatever"

    def test_attribute_not_found(self):
        field = fields.Raw()
        assert field.output("foo", {"bar": 42}) is None

    def test_object(self, mocker):
        obj = mocker.Mock()
        obj.foo = 42
        field = fields.Raw()
        assert field.output("foo", obj) == 42

    def test_nested_object(self, mocker):
        foo = mocker.Mock()
        bar = mocker.Mock()
        bar.value = 42
        foo.bar = bar
        field = fields.Raw()
        assert field.output("bar.value", foo) == 42


class StringFieldTest(StringTestMixin, BaseFieldTestMixin, FieldTestCase):
    field_class = fields.String

    def test_defaults(self):
        field = fields.String()
        assert not field.required
        assert not field.discriminator
        assert field.__schema__ == {"type": "string"}

    def test_with_enum(self):
        enum = ["A", "B", "C"]
        field = fields.String(enum=enum)
        assert not field.required
        assert field.__schema__ == {"type": "string", "enum": enum, "example": enum[0]}

    def test_with_empty_enum(self):
        field = fields.String(enum=[])
        assert not field.required
        assert field.__schema__ == {"type": "string"}

    def test_with_callable_enum(self):
        enum = lambda: ["A", "B", "C"]  # noqa
        field = fields.String(enum=enum)
        assert not field.required
        assert field.__schema__ == {
            "type": "string",
            "enum": ["A", "B", "C"],
            "example": "A",
        }

    def test_with_empty_callable_enum(self):
        enum = lambda: []  # noqa
        field = fields.String(enum=enum)
        assert not field.required
        assert field.__schema__ == {"type": "string"}

    def test_with_default(self):
        field = fields.String(default="aaa")
        assert field.__schema__ == {"type": "string", "default": "aaa"}

    def test_string_field_with_discriminator(self):
        field = fields.String(discriminator=True)
        assert field.discriminator
        assert field.required
        assert field.__schema__ == {"type": "string"}

    def test_string_field_with_discriminator_override_require(self):
        field = fields.String(discriminator=True, required=False)
        assert field.discriminator
        assert field.required
        assert field.__schema__ == {"type": "string"}

    def test_discriminator_output(self, api):
        model = api.model(
            "Test",
            {
                "name": fields.String(discriminator=True),
            },
        )

        data = api.marshal({}, model)
        assert data == {"name": "Test"}

    def test_multiple_discriminator_field(self, api):
        model = api.model(
            "Test",
            {
                "name": fields.String(discriminator=True),
                "name2": fields.String(discriminator=True),
            },
        )

        with pytest.raises(ValueError):
            api.marshal(object(), model)

    @pytest.mark.parametrize(
        "value,expected",
        [
            ("string", "string"),
            (42, "42"),
        ],
    )
    def test_values(self, value, expected):
        self.assert_field(fields.String(), value, expected)


class IntegerFieldTest(BaseFieldTestMixin, NumberTestMixin, FieldTestCase):
    field_class = fields.Integer

    def test_defaults(self):
        field = fields.Integer()
        assert not field.required
        assert field.__schema__ == {"type": "integer"}

    def test_with_default(self):
        field = fields.Integer(default=42)
        assert not field.required
        assert field.__schema__ == {"type": "integer", "default": 42}
        self.assert_field(field, None, 42)

    @pytest.mark.parametrize(
        "value,expected",
        [
            (0, 0),
            (42, 42),
            ("42", 42),
            (None, None),
            (66.6, 66),
        ],
    )
    def test_value(self, value, expected):
        self.assert_field(fields.Integer(), value, expected)

    def test_decode_error_on_invalid_value(self):
        field = fields.Integer()
        self.assert_field_raises(field, "an int")

    def test_decode_error_on_invalid_type(self):
        field = fields.Integer()
        self.assert_field_raises(field, {"a": "dict"})


class BooleanFieldTest(BaseFieldTestMixin, FieldTestCase):
    field_class = fields.Boolean

    def test_defaults(self):
        field = fields.Boolean()
        assert not field.required
        assert field.__schema__ == {"type": "boolean"}

    def test_with_default(self):
        field = fields.Boolean(default=True)
        assert not field.required
        assert field.__schema__ == {"type": "boolean", "default": True}

    def test_with_example(self):
        field = fields.Boolean(default=True, example=False)
        assert field.__schema__ == {
            "type": "boolean",
            "default": True,
            "example": False,
        }

    @pytest.mark.parametrize(
        "value,expected",
        [
            (True, True),
            (False, False),
            ({}, False),
            ("false", False),  # These consistent with inputs.boolean
            ("0", False),
        ],
    )
    def test_value(self, value, expected):
        self.assert_field(fields.Boolean(), value, expected)


class FloatFieldTest(BaseFieldTestMixin, NumberTestMixin, FieldTestCase):
    field_class = fields.Float

    def test_defaults(self):
        field = fields.Float()
        assert not field.required
        assert field.__schema__ == {"type": "number"}

    def test_with_default(self):
        field = fields.Float(default=0.5)
        assert not field.required
        assert field.__schema__ == {"type": "number", "default": 0.5}

    def test_none_uses_default(self):
        field = fields.Float(default=0.5)
        assert not field.required
        assert field.__schema__ == {"type": "number", "default": 0.5}
        assert field.format(None) == 0.5

    @pytest.mark.parametrize(
        "value,expected",
        [
            ("-3.13", -3.13),
            (str(-3.13), -3.13),
            (3, 3.0),
        ],
    )
    def test_value(self, value, expected):
        self.assert_field(fields.Float(), value, expected)

    def test_raises(self):
        self.assert_field_raises(fields.Float(), "bar")

    def test_decode_error_on_invalid_value(self):
        field = fields.Float()
        self.assert_field_raises(field, "not a float")

    def test_decode_error_on_invalid_type(self):
        field = fields.Float()
        self.assert_field_raises(field, {"a": "dict"})


PI_STR = (
    "3.141592653589793238462643383279502884197169399375105820974944592307816406286208998628034825342117"
    "06798214808651328230664709384460955058223172535940812848111745028410270193852110555964462294895493"
    "038196442881097566593344612847564823378678316527120190914564856692346034861"
)

PI = Decimal(PI_STR)


class FixedFieldTest(BaseFieldTestMixin, NumberTestMixin, FieldTestCase):
    field_class = fields.Fixed

    def test_defaults(self):
        field = fields.Fixed()
        assert not field.required
        assert field.__schema__ == {"type": "number"}

    def test_with_default(self):
        field = fields.Fixed(default=0.5)
        assert not field.required
        assert field.__schema__ == {"type": "number", "default": 0.5}

    def test_fixed(self):
        field5 = fields.Fixed(5)
        field4 = fields.Fixed(4)

        self.assert_field(field5, PI, "3.14159")
        self.assert_field(field4, PI, "3.1416")
        self.assert_field(field4, 3, "3.0000")
        self.assert_field(field4, "03", "3.0000")
        self.assert_field(field4, "03.0", "3.0000")

    def test_zero(self):
        self.assert_field(fields.Fixed(), "0", "0.00000")

    def test_infinite(self):
        field = fields.Fixed()
        self.assert_field_raises(field, "+inf")
        self.assert_field_raises(field, "-inf")

    def test_nan(self):
        field = fields.Fixed()
        self.assert_field_raises(field, "NaN")


class ArbitraryFieldTest(BaseFieldTestMixin, NumberTestMixin, FieldTestCase):
    field_class = fields.Arbitrary

    def test_defaults(self):
        field = fields.Arbitrary()
        assert not field.required
        assert field.__schema__ == {"type": "number"}

    def test_with_default(self):
        field = fields.Arbitrary(default=0.5)
        assert field.__schema__ == {"type": "number", "default": 0.5}

    @pytest.mark.parametrize(
        "value,expected",
        [
            (PI_STR, PI_STR),
            (PI, PI_STR),
        ],
    )
    def test_value(self, value, expected):
        self.assert_field(fields.Arbitrary(), value, expected)


class DatetimeFieldTest(BaseFieldTestMixin, FieldTestCase):
    field_class = fields.DateTime

    def test_defaults(self):
        field = fields.DateTime()
        assert not field.required
        assert field.__schema__ == {"type": "string", "format": "date-time"}
        self.assert_field(field, None, None)

    def test_with_default(self):
        field = fields.DateTime(default="2014-08-25")
        assert field.__schema__ == {
            "type": "string",
            "format": "date-time",
            "default": "2014-08-25T00:00:00",
        }
        self.assert_field(field, None, "2014-08-25T00:00:00")

    def test_with_default_as_datetime(self):
        field = fields.DateTime(default=datetime(2014, 8, 25))
        assert field.__schema__ == {
            "type": "string",
            "format": "date-time",
            "default": "2014-08-25T00:00:00",
        }
        self.assert_field(field, None, "2014-08-25T00:00:00")

    def test_with_default_as_date(self):
        field = fields.DateTime(default=date(2014, 8, 25))
        assert field.__schema__ == {
            "type": "string",
            "format": "date-time",
            "default": "2014-08-25T00:00:00",
        }
        self.assert_field(field, None, "2014-08-25T00:00:00")

    def test_min(self):
        field = fields.DateTime(min="1984-06-07T00:00:00")
        assert "minimum" in field.__schema__
        assert field.__schema__["minimum"] == "1984-06-07T00:00:00"
        assert "exclusiveMinimum" not in field.__schema__

    def test_min_as_date(self):
        field = fields.DateTime(min=date(1984, 6, 7))
        assert "minimum" in field.__schema__
        assert field.__schema__["minimum"] == "1984-06-07T00:00:00"
        assert "exclusiveMinimum" not in field.__schema__

    def test_min_as_datetime(self):
        field = fields.DateTime(min=datetime(1984, 6, 7, 1, 2, 0))
        assert "minimum" in field.__schema__
        assert field.__schema__["minimum"] == "1984-06-07T01:02:00"
        assert "exclusiveMinimum" not in field.__schema__

    def test_min_exlusive(self):
        field = fields.DateTime(min="1984-06-07T00:00:00", exclusiveMin=True)
        assert "minimum" in field.__schema__
        assert field.__schema__["minimum"] == "1984-06-07T00:00:00"
        assert "exclusiveMinimum" in field.__schema__
        assert field.__schema__["exclusiveMinimum"] is True

    def test_max(self):
        field = fields.DateTime(max="1984-06-07T00:00:00")
        assert "maximum" in field.__schema__
        assert field.__schema__["maximum"] == "1984-06-07T00:00:00"
        assert "exclusiveMaximum" not in field.__schema__

    def test_max_as_date(self):
        field = fields.DateTime(max=date(1984, 6, 7))
        assert "maximum" in field.__schema__
        assert field.__schema__["maximum"] == "1984-06-07T00:00:00"
        assert "exclusiveMaximum" not in field.__schema__

    def test_max_as_datetime(self):
        field = fields.DateTime(max=datetime(1984, 6, 7, 1, 2, 0))
        assert "maximum" in field.__schema__
        assert field.__schema__["maximum"] == "1984-06-07T01:02:00"
        assert "exclusiveMaximum" not in field.__schema__

    def test_max_exclusive(self):
        field = fields.DateTime(max="1984-06-07T00:00:00", exclusiveMax=True)
        assert "maximum" in field.__schema__
        assert field.__schema__["maximum"] == "1984-06-07T00:00:00"
        assert "exclusiveMaximum" in field.__schema__
        assert field.__schema__["exclusiveMaximum"] is True

    @pytest.mark.parametrize(
        "value,expected",
        [
            (date(2011, 1, 1), "Sat, 01 Jan 2011 00:00:00 -0000"),
            (datetime(2011, 1, 1), "Sat, 01 Jan 2011 00:00:00 -0000"),
            (datetime(2011, 1, 1, 23, 59, 59), "Sat, 01 Jan 2011 23:59:59 -0000"),
            (
                datetime(2011, 1, 1, 23, 59, 59, tzinfo=pytz.utc),
                "Sat, 01 Jan 2011 23:59:59 -0000",
            ),
            (
                datetime(2011, 1, 1, 23, 59, 59, tzinfo=pytz.timezone("CET")),
                "Sat, 01 Jan 2011 22:59:59 -0000",
            ),
        ],
    )
    def test_rfc822_value(self, value, expected):
        self.assert_field(fields.DateTime(dt_format="rfc822"), value, expected)

    @pytest.mark.parametrize(
        "value,expected",
        [
            (date(2011, 1, 1), "2011-01-01T00:00:00"),
            (datetime(2011, 1, 1), "2011-01-01T00:00:00"),
            (datetime(2011, 1, 1, 23, 59, 59), "2011-01-01T23:59:59"),
            (datetime(2011, 1, 1, 23, 59, 59, 1000), "2011-01-01T23:59:59.001000"),
            (
                datetime(2011, 1, 1, 23, 59, 59, tzinfo=pytz.utc),
                "2011-01-01T23:59:59+00:00",
            ),
            (
                datetime(2011, 1, 1, 23, 59, 59, 1000, tzinfo=pytz.utc),
                "2011-01-01T23:59:59.001000+00:00",
            ),
            (
                datetime(2011, 1, 1, 23, 59, 59, tzinfo=pytz.timezone("CET")),
                "2011-01-01T23:59:59+01:00",
            ),
        ],
    )
    def test_iso8601_value(self, value, expected):
        self.assert_field(fields.DateTime(dt_format="iso8601"), value, expected)

    def test_unsupported_format(self):
        field = fields.DateTime(dt_format="raw")
        self.assert_field_raises(field, datetime.now())

    def test_unsupported_value_format(self):
        field = fields.DateTime(dt_format="raw")
        self.assert_field_raises(field, "xxx")


class DateFieldTest(BaseFieldTestMixin, FieldTestCase):
    field_class = fields.Date

    def test_defaults(self):
        field = fields.Date()
        assert not field.required
        assert field.__schema__ == {"type": "string", "format": "date"}

    def test_with_default(self):
        field = fields.Date(default="2014-08-25")
        assert field.__schema__ == {
            "type": "string",
            "format": "date",
            "default": "2014-08-25",
        }
        self.assert_field(field, None, "2014-08-25")

    def test_with_default_as_date(self):
        field = fields.Date(default=date(2014, 8, 25))
        assert field.__schema__ == {
            "type": "string",
            "format": "date",
            "default": "2014-08-25",
        }

    def test_with_default_as_datetime(self):
        field = fields.Date(default=datetime(2014, 8, 25))
        assert field.__schema__ == {
            "type": "string",
            "format": "date",
            "default": "2014-08-25",
        }

    def test_min(self):
        field = fields.Date(min="1984-06-07")
        assert "minimum" in field.__schema__
        assert field.__schema__["minimum"] == "1984-06-07"
        assert "exclusiveMinimum" not in field.__schema__

    def test_min_as_date(self):
        field = fields.Date(min=date(1984, 6, 7))
        assert "minimum" in field.__schema__
        assert field.__schema__["minimum"] == "1984-06-07"
        assert "exclusiveMinimum" not in field.__schema__

    def test_min_as_datetime(self):
        field = fields.Date(min=datetime(1984, 6, 7, 1, 2, 0))
        assert "minimum" in field.__schema__
        assert field.__schema__["minimum"] == "1984-06-07"
        assert "exclusiveMinimum" not in field.__schema__

    def test_min_exlusive(self):
        field = fields.Date(min="1984-06-07", exclusiveMin=True)
        assert "minimum" in field.__schema__
        assert field.__schema__["minimum"] == "1984-06-07"
        assert "exclusiveMinimum" in field.__schema__
        assert field.__schema__["exclusiveMinimum"] is True

    def test_max(self):
        field = fields.Date(max="1984-06-07")
        assert "maximum" in field.__schema__
        assert field.__schema__["maximum"] == "1984-06-07"
        assert "exclusiveMaximum" not in field.__schema__

    def test_max_as_date(self):
        field = fields.Date(max=date(1984, 6, 7))
        assert "maximum" in field.__schema__
        assert field.__schema__["maximum"] == "1984-06-07"
        assert "exclusiveMaximum" not in field.__schema__

    def test_max_as_datetime(self):
        field = fields.Date(max=datetime(1984, 6, 7, 1, 2, 0))
        assert "maximum" in field.__schema__
        assert field.__schema__["maximum"] == "1984-06-07"
        assert "exclusiveMaximum" not in field.__schema__

    def test_max_exclusive(self):
        field = fields.Date(max="1984-06-07", exclusiveMax=True)
        assert "maximum" in field.__schema__
        assert field.__schema__["maximum"] == "1984-06-07"
        assert "exclusiveMaximum" in field.__schema__
        assert field.__schema__["exclusiveMaximum"] is True

    @pytest.mark.parametrize(
        "value,expected",
        [
            (date(2011, 1, 1), "2011-01-01"),
            (datetime(2011, 1, 1), "2011-01-01"),
            (datetime(2011, 1, 1, 23, 59, 59), "2011-01-01"),
            (datetime(2011, 1, 1, 23, 59, 59, 1000), "2011-01-01"),
            (datetime(2011, 1, 1, 23, 59, 59, tzinfo=pytz.utc), "2011-01-01"),
            (datetime(2011, 1, 1, 23, 59, 59, 1000, tzinfo=pytz.utc), "2011-01-01"),
            (
                datetime(2011, 1, 1, 23, 59, 59, tzinfo=pytz.timezone("CET")),
                "2011-01-01",
            ),
        ],
    )
    def test_value(self, value, expected):
        self.assert_field(fields.Date(), value, expected)

    def test_unsupported_value_format(self):
        self.assert_field_raises(fields.Date(), "xxx")


class FormatedStringFieldTest(StringTestMixin, BaseFieldTestMixin, FieldTestCase):
    field_class = partial(fields.FormattedString, "Hello {name}")

    def test_defaults(self):
        field = fields.FormattedString("Hello {name}")
        assert not field.required
        assert field.__schema__ == {"type": "string"}

    def test_dict(self):
        data = {
            "sid": 3,
            "account_sid": 4,
        }
        field = fields.FormattedString("/foo/{account_sid}/{sid}/")
        assert field.output("foo", data) == "/foo/4/3/"

    def test_object(self, mocker):
        obj = mocker.Mock()
        obj.sid = 3
        obj.account_sid = 4
        field = fields.FormattedString("/foo/{account_sid}/{sid}/")
        assert field.output("foo", obj) == "/foo/4/3/"

    def test_none(self):
        field = fields.FormattedString("{foo}")
        # self.assert_field_raises(field, None)
        with pytest.raises(fields.MarshallingError):
            field.output("foo", None)

    def test_invalid_object(self):
        field = fields.FormattedString("/foo/{0[account_sid]}/{0[sid]}/")
        self.assert_field_raises(field, {})

    def test_tuple(self):
        field = fields.FormattedString("/foo/{0[account_sid]}/{0[sid]}/")
        self.assert_field_raises(field, (3, 4))


class UrlFieldTest(StringTestMixin, BaseFieldTestMixin, FieldTestCase):
    field_class = partial(fields.Url, "endpoint")

    def test_defaults(self):
        field = fields.Url("endpoint")
        assert not field.required
        assert field.__schema__ == {"type": "string"}

    def test_invalid_object(self, app):
        app.add_url_rule("/<foo>", "foobar", view_func=lambda x: x)
        field = fields.Url("foobar")

        with app.test_request_context("/"):
            with pytest.raises(fields.MarshallingError):
                field.output("foo", None)

    def test_simple(self, app, mocker):
        app.add_url_rule("/<foo>", "foobar", view_func=lambda x: x)
        field = fields.Url("foobar")
        obj = mocker.Mock(foo=42)

        with app.test_request_context("/"):
            assert "/42" == field.output("foo", obj)

    def test_absolute(self, app, mocker):
        app.add_url_rule("/<foo>", "foobar", view_func=lambda x: x)
        field = fields.Url("foobar", absolute=True)
        obj = mocker.Mock(foo=42)

        with app.test_request_context("/"):
            assert "http://localhost/42" == field.output("foo", obj)

    def test_absolute_scheme(self, app, mocker):
        """Url.scheme should override current_request.scheme"""
        app.add_url_rule("/<foo>", "foobar", view_func=lambda x: x)
        field = fields.Url("foobar", absolute=True, scheme="https")
        obj = mocker.Mock(foo=42)

        with app.test_request_context("/", base_url="http://localhost"):
            assert "https://localhost/42" == field.output("foo", obj)

    def test_without_endpoint_invalid_object(self, app):
        app.add_url_rule("/<foo>", "foobar", view_func=lambda x: x)
        field = fields.Url()

        with app.test_request_context("/foo"):
            with pytest.raises(fields.MarshallingError):
                field.output("foo", None)

    def test_without_endpoint(self, app, mocker):
        app.add_url_rule("/<foo>", "foobar", view_func=lambda x: x)
        field = fields.Url()
        obj = mocker.Mock(foo=42)

        with app.test_request_context("/foo"):
            assert "/42" == field.output("foo", obj)

    def test_without_endpoint_absolute(self, app, mocker):
        app.add_url_rule("/<foo>", "foobar", view_func=lambda x: x)
        field = fields.Url(absolute=True)
        obj = mocker.Mock(foo=42)

        with app.test_request_context("/foo"):
            assert "http://localhost/42" == field.output("foo", obj)

    def test_without_endpoint_absolute_scheme(self, app, mocker):
        app.add_url_rule("/<foo>", "foobar", view_func=lambda x: x)
        field = fields.Url(absolute=True, scheme="https")
        obj = mocker.Mock(foo=42)

        with app.test_request_context("/foo", base_url="http://localhost"):
            assert "https://localhost/42" == field.output("foo", obj)

    def test_with_blueprint_invalid_object(self, app):
        bp = Blueprint("foo", __name__, url_prefix="/foo")
        bp.add_url_rule("/<foo>", "foobar", view_func=lambda x: x)
        app.register_blueprint(bp)
        field = fields.Url()

        with app.test_request_context("/foo/foo"):
            with pytest.raises(fields.MarshallingError):
                field.output("foo", None)

    def test_with_blueprint(self, app, mocker):
        bp = Blueprint("foo", __name__, url_prefix="/foo")
        bp.add_url_rule("/<foo>", "foobar", view_func=lambda x: x)
        app.register_blueprint(bp)
        field = fields.Url()
        obj = mocker.Mock(foo=42)

        with app.test_request_context("/foo/foo"):
            assert "/foo/42" == field.output("foo", obj)

    def test_with_blueprint_absolute(self, app, mocker):
        bp = Blueprint("foo", __name__, url_prefix="/foo")
        bp.add_url_rule("/<foo>", "foobar", view_func=lambda x: x)
        app.register_blueprint(bp)
        field = fields.Url(absolute=True)
        obj = mocker.Mock(foo=42)

        with app.test_request_context("/foo/foo"):
            assert "http://localhost/foo/42" == field.output("foo", obj)

    def test_with_blueprint_absolute_scheme(self, app, mocker):
        bp = Blueprint("foo", __name__, url_prefix="/foo")
        bp.add_url_rule("/<foo>", "foobar", view_func=lambda x: x)
        app.register_blueprint(bp)
        field = fields.Url(absolute=True, scheme="https")
        obj = mocker.Mock(foo=42)

        with app.test_request_context("/foo/foo", base_url="http://localhost"):
            assert "https://localhost/foo/42" == field.output("foo", obj)


class NestedFieldTest(FieldTestCase):
    def test_defaults(self, api):
        nested_fields = api.model("NestedModel", {"name": fields.String})
        field = fields.Nested(nested_fields)
        assert not field.required
        assert field.__schema__ == {"$ref": "#/definitions/NestedModel"}

    def test_with_required(self, api):
        nested_fields = api.model("NestedModel", {"name": fields.String})
        field = fields.Nested(nested_fields, required=True)
        assert field.required
        assert not field.allow_null
        assert field.__schema__ == {"$ref": "#/definitions/NestedModel"}

    def test_with_description(self, api):
        nested_fields = api.model("NestedModel", {"name": fields.String})
        field = fields.Nested(nested_fields, description="A description")
        assert field.__schema__ == {
            "description": "A description",
            "allOf": [{"$ref": "#/definitions/NestedModel"}],
        }

    def test_with_title(self, api):
        nested_fields = api.model("NestedModel", {"name": fields.String})
        field = fields.Nested(nested_fields, title="A title")
        assert field.__schema__ == {
            "title": "A title",
            "allOf": [{"$ref": "#/definitions/NestedModel"}],
        }

    def test_with_allow_null(self, api):
        nested_fields = api.model("NestedModel", {"name": fields.String})
        field = fields.Nested(nested_fields, allow_null=True)
        assert not field.required
        assert field.allow_null
        assert field.__schema__ == {"$ref": "#/definitions/NestedModel"}

    def test_with_skip_none(self, api):
        nested_fields = api.model("NestedModel", {"name": fields.String})
        field = fields.Nested(nested_fields, skip_none=True)
        assert not field.required
        assert field.skip_none
        assert field.__schema__ == {"$ref": "#/definitions/NestedModel"}

    def test_with_readonly(self, app):
        api = Api(app)
        nested_fields = api.model("NestedModel", {"name": fields.String})
        field = fields.Nested(nested_fields, readonly=True)
        assert field.__schema__ == {
            "readOnly": True,
            "allOf": [{"$ref": "#/definitions/NestedModel"}],
        }

    def test_as_list(self, api):
        nested_fields = api.model("NestedModel", {"name": fields.String})
        field = fields.Nested(nested_fields, as_list=True)
        assert field.as_list
        assert field.__schema__ == {
            "type": "array",
            "items": {"$ref": "#/definitions/NestedModel"},
        }

    def test_as_list_is_reusable(self, api):
        nested_fields = api.model("NestedModel", {"name": fields.String})

        field = fields.Nested(nested_fields, as_list=True)
        assert field.__schema__ == {
            "type": "array",
            "items": {"$ref": "#/definitions/NestedModel"},
        }

        field = fields.Nested(nested_fields)
        assert field.__schema__ == {"$ref": "#/definitions/NestedModel"}


class ListFieldTest(BaseFieldTestMixin, FieldTestCase):
    field_class = partial(fields.List, fields.String)

    def test_defaults(self):
        field = fields.List(fields.String)
        assert not field.required
        assert field.__schema__ == {"type": "array", "items": {"type": "string"}}

    def test_with_nested_field(self, api):
        nested_fields = api.model("NestedModel", {"name": fields.String})
        field = fields.List(fields.Nested(nested_fields))
        assert field.__schema__ == {
            "type": "array",
            "items": {"$ref": "#/definitions/NestedModel"},
        }

        data = [{"name": "John Doe", "age": 42}, {"name": "Jane Doe", "age": 66}]
        expected = [
            OrderedDict([("name", "John Doe")]),
            OrderedDict([("name", "Jane Doe")]),
        ]
        self.assert_field(field, data, expected)

    def test_min_items(self):
        field = fields.List(fields.String, min_items=5)
        assert "minItems" in field.__schema__
        assert field.__schema__["minItems"] == 5

    def test_max_items(self):
        field = fields.List(fields.String, max_items=42)
        assert "maxItems" in field.__schema__
        assert field.__schema__["maxItems"] == 42

    def test_unique(self):
        field = fields.List(fields.String, unique=True)
        assert "uniqueItems" in field.__schema__
        assert field.__schema__["uniqueItems"] is True

    @pytest.mark.parametrize(
        "value,expected",
        [
            (["a", "b", "c"], ["a", "b", "c"]),
            (["c", "b", "a"], ["c", "b", "a"]),
            (("a", "b", "c"), ["a", "b", "c"]),
            (["a"], ["a"]),
            (None, None),
        ],
    )
    def test_value(self, value, expected):
        self.assert_field(fields.List(fields.String()), value, expected)

    def test_with_set(self):
        field = fields.List(fields.String)
        value = set(["a", "b", "c"])
        output = field.output("foo", {"foo": value})
        assert set(output) == value

    def test_with_scoped_attribute_on_dict_or_obj(self):
        class Test(object):
            def __init__(self, data):
                self.data = data

        class Nested(object):
            def __init__(self, value):
                self.value = value

        nesteds = [Nested(i) for i in ["a", "b", "c"]]
        test_obj = Test(nesteds)
        test_dict = {"data": [{"value": "a"}, {"value": "b"}, {"value": "c"}]}

        field = fields.List(fields.String(attribute="value"), attribute="data")
        assert ["a" == "b", "c"], field.output("whatever", test_obj)
        assert ["a" == "b", "c"], field.output("whatever", test_dict)

    def test_with_attribute(self):
        data = [{"a": 1, "b": 1}, {"a": 2, "b": 1}, {"a": 3, "b": 1}]
        field = fields.List(fields.Integer(attribute="a"))
        self.assert_field(field, data, [1, 2, 3])

    def test_list_of_raw(self):
        field = fields.List(fields.Raw)

        data = [{"a": 1, "b": 1}, {"a": 2, "b": 1}, {"a": 3, "b": 1}]
        expected = [
            OrderedDict([("a", 1), ("b", 1)]),
            OrderedDict([("a", 2), ("b", 1)]),
            OrderedDict([("a", 3), ("b", 1)]),
        ]
        self.assert_field(field, data, expected)

        data = [1, 2, "a"]
        self.assert_field(field, data, data)


class WildcardFieldTest(BaseFieldTestMixin, FieldTestCase):
    field_class = partial(fields.Wildcard, fields.String)

    def test_types(self):
        with pytest.raises(fields.MarshallingError):

            class WrongType:
                pass

            x = WrongType()
            field1 = fields.Wildcard(WrongType)  # noqa
            field2 = fields.Wildcard(x)  # noqa

    def test_defaults(self):
        field = fields.Wildcard(fields.String)
        assert not field.required
        assert field.__schema__ == {
            "type": "object",
            "additionalProperties": {"type": "string"},
        }

    def test_with_scoped_attribute_on_dict_or_obj(self):
        class Test(object):
            def __init__(self, data):
                self.data = data

        class Nested(object):
            def __init__(self, value):
                self.value = value

        nesteds = [Nested(i) for i in ["a", "b", "c"]]
        test_obj = Test(nesteds)
        test_dict = {"data": [{"value": "a"}, {"value": "b"}, {"value": "c"}]}

        field = fields.Wildcard(fields.String(attribute="value"), attribute="data")
        assert ["a" == "b", "c"], field.output("whatever", test_obj)
        assert ["a" == "b", "c"], field.output("whatever", test_dict)

    def test_list_of_raw(self):
        field = fields.Wildcard(fields.Raw)

        data = [{"a": 1, "b": 1}, {"a": 2, "b": 1}, {"a": 3, "b": 1}]
        expected = [
            OrderedDict([("a", 1), ("b", 1)]),
            OrderedDict([("a", 2), ("b", 1)]),
            OrderedDict([("a", 3), ("b", 1)]),
        ]
        self.assert_field(field, data, expected)

        data = [1, 2, "a"]
        self.assert_field(field, data, data)

    def test_wildcard(self, api):
        wild1 = fields.Wildcard(fields.String)
        wild2 = fields.Wildcard(fields.Integer)
        wild3 = fields.Wildcard(fields.String)
        wild4 = fields.Wildcard(fields.String, default="x")
        wild5 = fields.Wildcard(fields.String)
        wild6 = fields.Wildcard(fields.Integer)
        wild7 = fields.Wildcard(fields.String)
        wild8 = fields.Wildcard(fields.String)

        mod5 = OrderedDict()
        mod5["toto"] = fields.Integer
        mod5["bob"] = fields.Integer
        mod5["*"] = wild5

        wild_fields1 = api.model("WildcardModel1", {"*": wild1})
        wild_fields2 = api.model("WildcardModel2", {"j*": wild2})
        wild_fields3 = api.model("WildcardModel3", {"*": wild3})
        wild_fields4 = api.model("WildcardModel4", {"*": wild4})
        wild_fields5 = api.model("WildcardModel5", mod5)
        wild_fields6 = api.model(
            "WildcardModel6",
            {
                "nested": {
                    "f1": fields.String(default="12"),
                    "f2": fields.Integer(default=13),
                },
                "a*": wild6,
            },
        )
        wild_fields7 = api.model("WildcardModel7", {"*": wild7})
        wild_fields8 = api.model("WildcardModel8", {"*": wild8})

        class Dummy(object):
            john = 12
            bob = "42"
            alice = None

        class Dummy2(object):
            pass

        class Dummy3(object):
            a = None
            b = None

        data = {"John": 12, "bob": 42, "Jane": "68"}
        data3 = Dummy()
        data4 = Dummy2()
        data5 = {"John": 12, "bob": 42, "Jane": "68", "toto": "72"}
        data6 = {"nested": {"f1": 12, "f2": 13}, "alice": "14"}
        data7 = Dummy3()
        data8 = None
        expected1 = {"John": "12", "bob": "42", "Jane": "68"}
        expected2 = {"John": 12, "Jane": 68}
        expected3 = {"john": "12", "bob": "42"}
        expected4 = {"*": "x"}
        expected5 = {"John": "12", "bob": 42, "Jane": "68", "toto": 72}
        expected6 = {"nested": {"f1": "12", "f2": 13}, "alice": 14}
        expected7 = {}
        expected8 = {}

        result1 = api.marshal(data, wild_fields1)
        result2 = api.marshal(data, wild_fields2)
        result3 = api.marshal(data3, wild_fields3, skip_none=True)
        result4 = api.marshal(data4, wild_fields4)
        result5 = api.marshal(data5, wild_fields5)
        result6 = api.marshal(data6, wild_fields6)
        result7 = api.marshal(data7, wild_fields7, skip_none=True)
        result8 = api.marshal(data8, wild_fields8, skip_none=True)

        assert expected1 == result1
        assert expected2 == result2
        assert expected3 == result3
        assert expected4 == result4
        assert expected5 == result5
        assert expected6 == result6
        assert expected7 == result7
        assert expected8 == result8

    def test_clone(self, api):
        wild1 = fields.Wildcard(fields.String)
        wild2 = wild1.clone()

        wild_fields1 = api.model("cloneWildcard1", {"*": wild1})
        wild_fields2 = api.model("cloneWildcard2", {"*": wild2})

        data = {"John": 12, "bob": 42, "Jane": "68"}
        expected1 = {"John": "12", "bob": "42", "Jane": "68"}

        result1 = api.marshal(data, wild_fields1)
        result2 = api.marshal(data, wild_fields2)

        assert expected1 == result1
        assert result2 == result1


class ClassNameFieldTest(StringTestMixin, BaseFieldTestMixin, FieldTestCase):
    field_class = fields.ClassName

    def test_simple_string_field(self):
        field = fields.ClassName()
        assert not field.required
        assert not field.discriminator
        assert field.__schema__ == {"type": "string"}

    def test_default_output_classname(self, api):
        model = api.model(
            "Test",
            {
                "name": fields.ClassName(),
            },
        )

        class FakeClass(object):
            pass

        data = api.marshal(FakeClass(), model)
        assert data == {"name": "FakeClass"}

    def test_output_dash(self, api):
        model = api.model(
            "Test",
            {
                "name": fields.ClassName(dash=True),
            },
        )

        class FakeClass(object):
            pass

        data = api.marshal(FakeClass(), model)
        assert data == {"name": "fake_class"}

    def test_with_dict(self, api):
        model = api.model(
            "Test",
            {
                "name": fields.ClassName(),
            },
        )

        data = api.marshal({}, model)
        assert data == {"name": "object"}


class PolymorphTest(FieldTestCase):
    def test_polymorph_field(self, api):
        parent = api.model(
            "Person",
            {
                "name": fields.String,
            },
        )

        child1 = api.inherit(
            "Child1",
            parent,
            {
                "extra1": fields.String,
            },
        )

        child2 = api.inherit(
            "Child2",
            parent,
            {
                "extra2": fields.String,
            },
        )

        class Child1(object):
            name = "child1"
            extra1 = "extra1"

        class Child2(object):
            name = "child2"
            extra2 = "extra2"

        mapping = {Child1: child1, Child2: child2}

        thing = api.model(
            "Thing",
            {
                "owner": fields.Polymorph(mapping),
            },
        )

        def data(cls):
            return api.marshal({"owner": cls()}, thing)

        assert data(Child1) == {"owner": {"name": "child1", "extra1": "extra1"}}

        assert data(Child2) == {"owner": {"name": "child2", "extra2": "extra2"}}

    def test_polymorph_field_no_common_ancestor(self, api):
        child1 = api.model(
            "Child1",
            {
                "extra1": fields.String,
            },
        )

        child2 = api.model(
            "Child2",
            {
                "extra2": fields.String,
            },
        )

        class Child1(object):
            pass

        class Child2(object):
            pass

        mapping = {Child1: child1, Child2: child2}

        with pytest.raises(ValueError):
            fields.Polymorph(mapping)

    def test_polymorph_field_unknown_class(self, api):
        parent = api.model(
            "Person",
            {
                "name": fields.String,
            },
        )

        child1 = api.inherit(
            "Child1",
            parent,
            {
                "extra1": fields.String,
            },
        )

        child2 = api.inherit(
            "Child2",
            parent,
            {
                "extra2": fields.String,
            },
        )

        class Child1(object):
            name = "child1"
            extra1 = "extra1"

        class Child2(object):
            name = "child2"
            extra2 = "extra2"

        mapping = {Child1: child1, Child2: child2}

        thing = api.model(
            "Thing",
            {
                "owner": fields.Polymorph(mapping),
            },
        )

        with pytest.raises(ValueError):
            api.marshal({"owner": object()}, thing)

    def test_polymorph_field_does_not_have_ambiguous_mappings(self, api):
        """
        Regression test for https://github.com/noirbizarre/flask-restx/pull/691
        """
        parent = api.model(
            "Parent",
            {
                "name": fields.String,
            },
        )

        child = api.inherit(
            "Child",
            parent,
            {
                "extra": fields.String,
            },
        )

        class Parent(object):
            name = "parent"

        class Child(Parent):
            extra = "extra"

        mapping = {Parent: parent, Child: child}

        thing = api.model(
            "Thing",
            {
                "owner": fields.Polymorph(mapping),
            },
        )

        api.marshal({"owner": Child()}, thing)

    def test_polymorph_field_required_default(self, api):
        parent = api.model(
            "Person",
            {
                "name": fields.String,
            },
        )

        child1 = api.inherit(
            "Child1",
            parent,
            {
                "extra1": fields.String,
            },
        )

        child2 = api.inherit(
            "Child2",
            parent,
            {
                "extra2": fields.String,
            },
        )

        class Child1(object):
            name = "child1"
            extra1 = "extra1"

        class Child2(object):
            name = "child2"
            extra2 = "extra2"

        mapping = {Child1: child1, Child2: child2}

        thing = api.model(
            "Thing",
            {
                "owner": fields.Polymorph(
                    mapping, required=True, default={"name": "default"}
                ),
            },
        )

        data = api.marshal({}, thing)

        assert data == {"owner": {"name": "default"}}

    def test_polymorph_field_not_required(self, api):
        parent = api.model(
            "Person",
            {
                "name": fields.String,
            },
        )

        child1 = api.inherit(
            "Child1",
            parent,
            {
                "extra1": fields.String,
            },
        )

        child2 = api.inherit(
            "Child2",
            parent,
            {
                "extra2": fields.String,
            },
        )

        class Child1(object):
            name = "child1"
            extra1 = "extra1"

        class Child2(object):
            name = "child2"
            extra2 = "extra2"

        mapping = {Child1: child1, Child2: child2}

        thing = api.model(
            "Thing",
            {
                "owner": fields.Polymorph(mapping),
            },
        )

        data = api.marshal({}, thing)

        assert data == {"owner": None}

    def test_polymorph_with_discriminator(self, api):
        parent = api.model(
            "Person",
            {
                "name": fields.String,
                "model": fields.String(discriminator=True),
            },
        )

        child1 = api.inherit(
            "Child1",
            parent,
            {
                "extra1": fields.String,
            },
        )

        child2 = api.inherit(
            "Child2",
            parent,
            {
                "extra2": fields.String,
            },
        )

        class Child1(object):
            name = "child1"
            extra1 = "extra1"

        class Child2(object):
            name = "child2"
            extra2 = "extra2"

        mapping = {Child1: child1, Child2: child2}

        thing = api.model(
            "Thing",
            {
                "owner": fields.Polymorph(mapping),
            },
        )

        def data(cls):
            return api.marshal({"owner": cls()}, thing)

        assert data(Child1) == {
            "owner": {"name": "child1", "model": "Child1", "extra1": "extra1"}
        }

        assert data(Child2) == {
            "owner": {"name": "child2", "model": "Child2", "extra2": "extra2"}
        }


class CustomFieldTest(FieldTestCase):
    def test_custom_field(self):
        class CustomField(fields.Integer):
            __schema_format__ = "int64"

        field = CustomField()

        assert field.__schema__ == {"type": "integer", "format": "int64"}


class FieldsHelpersTest(object):
    def test_to_dict(self):
        expected = data = {"foo": 42}
        assert fields.to_marshallable_type(data) == expected

    def test_to_dict_obj(self):
        class Foo(object):
            def __init__(self):
                self.foo = 42

        expected = {"foo": 42}
        assert fields.to_marshallable_type(Foo()) == expected

    def test_to_dict_custom_marshal(self):
        class Foo(object):
            def __marshallable__(self):
                return {"foo": 42}

        expected = {"foo": 42}
        assert fields.to_marshallable_type(Foo()) == expected

    def test_get_value(self):
        assert fields.get_value("foo", {"foo": 42}) == 42

    def test_get_value_no_value(self):
        assert fields.get_value("foo", {"foo": 42}) == 42

    def test_get_value_obj(self, mocker):
        assert fields.get_value("foo", mocker.Mock(foo=42)) == 42

    def test_get_value_indexable_object(self):
        class Test(object):
            def __init__(self, value):
                self.value = value

            def __getitem__(self, n):
                if type(n) is int:
                    if n < 3:
                        return n
                    raise IndexError
                raise TypeError

        obj = Test("hi")
        assert fields.get_value("value", obj) == "hi"

    def test_get_value_int_indexable_list(self):
        assert fields.get_value("bar.0", {"bar": [42]}) == 42

    def test_get_value_int_indexable_list_with_str(self):
        assert fields.get_value("bar.abc", {"bar": [42]}) is None

    def test_get_value_int_indexable_nested_list(self):
        assert fields.get_value("bar.0.val", {"bar": [{"val": 42}]}) == 42

    def test_get_value_int_indexable_tuple_with_str(self):
        assert fields.get_value("bar.abc", {"bar": (42, 43)}) is None

    def test_get_value_int_indexable_tuple(self):
        assert fields.get_value("bar.0", {"bar": (42, 43)}) == 42

    def test_get_value_int_indexable_nested_tuple(self):
        assert fields.get_value("bar.0.val", {"bar": [{"val": 42}]}) == 42



import json
import logging

import pytest

from flask import Blueprint, abort
from flask.signals import got_request_exception

from werkzeug import Response
from werkzeug.exceptions import (
    Aborter,
    BadRequest,
    HTTPException,
    NotFound,
    Unauthorized,
)
from werkzeug.http import quote_etag, unquote_etag

import flask_restx as restx


class ErrorsTest(object):
    def test_abort_type(self):
        with pytest.raises(HTTPException):
            restx.abort(404)

    def test_abort_data(self):
        with pytest.raises(HTTPException) as cm:
            restx.abort(404, foo="bar")
        assert cm.value.data == {"foo": "bar"}

    def test_abort_no_data(self):
        with pytest.raises(HTTPException) as cm:
            restx.abort(404)
        assert not hasattr(cm.value, "data")

    def test_abort_custom_message(self):
        with pytest.raises(HTTPException) as cm:
            restx.abort(404, "My message")
        assert cm.value.data["message"] == "My message"

    def test_abort_code_only_with_defaults(self, app, client):
        api = restx.Api(app)

        @api.route("/test/", endpoint="test")
        class TestResource(restx.Resource):
            def get(self):
                api.abort(403)

        response = client.get("/test/")
        assert response.status_code == 403
        assert response.content_type == "application/json"

        data = json.loads(response.data.decode("utf8"))
        assert "message" in data

    def test_abort_with_message(self, app, client):
        api = restx.Api(app)

        @api.route("/test/", endpoint="test")
        class TestResource(restx.Resource):
            def get(self):
                api.abort(403, "A message")

        response = client.get("/test/")
        assert response.status_code == 403
        assert response.content_type == "application/json"

        data = json.loads(response.data.decode("utf8"))
        assert data["message"] == "A message"

    def test_abort_with_lazy_init(self, app, client):
        api = restx.Api()

        @api.route("/test/", endpoint="test")
        class TestResource(restx.Resource):
            def get(self):
                api.abort(403)

        api.init_app(app)

        response = client.get("/test/")
        assert response.status_code == 403
        assert response.content_type == "application/json"

        data = json.loads(response.data.decode("utf8"))
        assert "message" in data

    def test_abort_on_exception(self, app, client):
        api = restx.Api(app)

        @api.route("/test/", endpoint="test")
        class TestResource(restx.Resource):
            def get(self):
                raise ValueError()

        response = client.get("/test/")
        assert response.status_code == 500
        assert response.content_type == "application/json"

        data = json.loads(response.data.decode("utf8"))
        assert "message" in data

    def test_abort_on_exception_with_lazy_init(self, app, client):
        api = restx.Api()

        @api.route("/test/", endpoint="test")
        class TestResource(restx.Resource):
            def get(self):
                raise ValueError()

        api.init_app(app)

        response = client.get("/test/")
        assert response.status_code == 500
        assert response.content_type == "application/json"

        data = json.loads(response.data.decode("utf8"))
        assert "message" in data

    def test_errorhandler_for_exception_inheritance(self, app, client):
        api = restx.Api(app)

        class CustomException(RuntimeError):
            pass

        @api.route("/test/", endpoint="test")
        class TestResource(restx.Resource):
            def get(self):
                raise CustomException("error")

        @api.errorhandler(RuntimeError)
        def handle_custom_exception(error):
            return {"message": str(error), "test": "value"}, 400

        response = client.get("/test/")
        assert response.status_code == 400
        assert response.content_type == "application/json"

        data = json.loads(response.data.decode("utf8"))
        assert data == {
            "message": "error",
            "test": "value",
        }

    def test_errorhandler_for_custom_exception(self, app, client):
        api = restx.Api(app)

        class CustomException(RuntimeError):
            pass

        @api.route("/test/", endpoint="test")
        class TestResource(restx.Resource):
            def get(self):
                raise CustomException("error")

        @api.errorhandler(CustomException)
        def handle_custom_exception(error):
            return {"message": str(error), "test": "value"}, 400

        response = client.get("/test/")
        assert response.status_code == 400
        assert response.content_type == "application/json"

        data = json.loads(response.data.decode("utf8"))
        assert data == {
            "message": "error",
            "test": "value",
        }

    def test_blunder_in_errorhandler_is_not_suppressed_in_logs(
        self, app, client, caplog
    ):
        api = restx.Api(app)

        class CustomException(RuntimeError):
            pass

        class ProgrammingBlunder(Exception):
            pass

        @api.route("/test/", endpoint="test")
        class TestResource(restx.Resource):
            def get(self):
                raise CustomException("error")

        @api.errorhandler(CustomException)
        def handle_custom_exception(error):
            raise ProgrammingBlunder(
                "This exception needs to be logged, not suppressed, then cause 500"
            )

        with caplog.at_level(logging.ERROR):
            response = client.get("/test/")
        exc_type, value, traceback = caplog.records[0].exc_info
        assert exc_type is ProgrammingBlunder
        assert response.status_code == 500

    def test_errorhandler_for_custom_exception_with_headers(self, app, client):
        api = restx.Api(app)

        class CustomException(RuntimeError):
            pass

        @api.route("/test/", endpoint="test")
        class TestResource(restx.Resource):
            def get(self):
                raise CustomException("error")

        @api.errorhandler(CustomException)
        def handle_custom_exception(error):
            return {"message": "some maintenance"}, 503, {"Retry-After": 120}

        response = client.get("/test/")
        assert response.status_code == 503
        assert response.content_type == "application/json"

        data = json.loads(response.data.decode("utf8"))
        assert data == {"message": "some maintenance"}
        assert response.headers["Retry-After"] == "120"

    def test_errorhandler_for_httpexception(self, app, client):
        api = restx.Api(app)

        @api.route("/test/", endpoint="test")
        class TestResource(restx.Resource):
            def get(self):
                raise BadRequest()

        @api.errorhandler(BadRequest)
        def handle_badrequest_exception(error):
            return {"message": str(error), "test": "value"}, 400

        response = client.get("/test/")
        assert response.status_code == 400
        assert response.content_type == "application/json"

        data = json.loads(response.data.decode("utf8"))
        assert data == {
            "message": str(BadRequest()),
            "test": "value",
        }

    def test_errorhandler_with_namespace(self, app, client):
        api = restx.Api(app)

        ns = restx.Namespace("ExceptionHandler", path="/")

        class CustomException(RuntimeError):
            pass

        @ns.route("/test/", endpoint="test")
        class TestResource(restx.Resource):
            def get(self):
                raise CustomException("error")

        @ns.errorhandler(CustomException)
        def handle_custom_exception(error):
            return {"message": str(error), "test": "value"}, 400

        api.add_namespace(ns)

        response = client.get("/test/")
        assert response.status_code == 400
        assert response.content_type == "application/json"

        data = json.loads(response.data.decode("utf8"))
        assert data == {
            "message": "error",
            "test": "value",
        }

    def test_errorhandler_with_namespace_from_api(self, app, client):
        api = restx.Api(app)

        ns = api.namespace("ExceptionHandler", path="/")

        class CustomException(RuntimeError):
            pass

        @ns.route("/test/", endpoint="test")
        class TestResource(restx.Resource):
            def get(self):
                raise CustomException("error")

        @ns.errorhandler(CustomException)
        def handle_custom_exception(error):
            return {"message": str(error), "test": "value"}, 400

        response = client.get("/test/")
        assert response.status_code == 400
        assert response.content_type == "application/json"

        data = json.loads(response.data.decode("utf8"))
        assert data == {
            "message": "error",
            "test": "value",
        }

    def test_default_errorhandler(self, app, client):
        api = restx.Api(app)

        @api.route("/test/")
        class TestResource(restx.Resource):
            def get(self):
                raise Exception("error")

        response = client.get("/test/")
        assert response.status_code == 500
        assert response.content_type == "application/json"

        data = json.loads(response.data.decode("utf8"))
        assert "message" in data

    def test_default_errorhandler_with_propagate_true(self, app, client):
        blueprint = Blueprint("api", __name__, url_prefix="/api")
        api = restx.Api(blueprint)

        @api.route("/test/")
        class TestResource(restx.Resource):
            def get(self):
                raise Exception("error")

        app.register_blueprint(blueprint)

        app.config["PROPAGATE_EXCEPTIONS"] = True

        # From the Flask docs:
        # PROPAGATE_EXCEPTIONS
        # Exceptions are re-raised rather than being handled by the app’s error handlers.
        # If not set, this is implicitly true if TESTING or DEBUG is enabled.
        with pytest.raises(Exception):
            client.get("/api/test/")

    def test_default_errorhandler_with_propagate_not_set_but_testing(self, app, client):
        blueprint = Blueprint("api", __name__, url_prefix="/api")
        api = restx.Api(blueprint)

        @api.route("/test/")
        class TestResource(restx.Resource):
            def get(self):
                raise Exception("error")

        app.register_blueprint(blueprint)

        app.config["PROPAGATE_EXCEPTIONS"] = None
        app.testing = True

        # From the Flask docs:
        # PROPAGATE_EXCEPTIONS
        # Exceptions are re-raised rather than being handled by the app’s error handlers.
        # If not set, this is implicitly true if TESTING or DEBUG is enabled.
        with pytest.raises(Exception):
            client.get("/api/test/")

    def test_default_errorhandler_with_propagate_not_set_but_debug(self, app, client):
        blueprint = Blueprint("api", __name__, url_prefix="/api")
        api = restx.Api(blueprint)

        @api.route("/test/")
        class TestResource(restx.Resource):
            def get(self):
                raise Exception("error")

        app.register_blueprint(blueprint)

        app.config["PROPAGATE_EXCEPTIONS"] = None
        app.debug = True

        # From the Flask docs:
        # PROPAGATE_EXCEPTIONS
        # Exceptions are re-raised rather than being handled by the app’s error handlers.
        # If not set, this is implicitly true if TESTING or DEBUG is enabled.
        with pytest.raises(Exception):
            client.get("/api/test/")

    def test_custom_default_errorhandler(self, app, client):
        api = restx.Api(app)

        @api.route("/test/", endpoint="test")
        class TestResource(restx.Resource):
            def get(self):
                raise Exception("error")

        @api.errorhandler
        def default_error_handler(error):
            return {"message": str(error), "test": "value"}, 500

        response = client.get("/test/")
        assert response.status_code == 500
        assert response.content_type == "application/json"

        data = json.loads(response.data.decode("utf8"))
        assert data == {
            "message": "error",
            "test": "value",
        }

    def test_custom_default_errorhandler_with_headers(self, app, client):
        api = restx.Api(app)

        @api.route("/test/", endpoint="test")
        class TestResource(restx.Resource):
            def get(self):
                raise Exception("error")

        @api.errorhandler
        def default_error_handler(error):
            return {"message": "some maintenance"}, 503, {"Retry-After": 120}

        response = client.get("/test/")
        assert response.status_code == 503
        assert response.content_type == "application/json"

        data = json.loads(response.data.decode("utf8"))
        assert data == {"message": "some maintenance"}
        assert response.headers["Retry-After"] == "120"

    def test_errorhandler_lazy(self, app, client):
        api = restx.Api()

        class CustomException(RuntimeError):
            pass

        @api.route("/test/", endpoint="test")
        class TestResource(restx.Resource):
            def get(self):
                raise CustomException("error")

        @api.errorhandler(CustomException)
        def handle_custom_exception(error):
            return {"message": str(error), "test": "value"}, 400

        api.init_app(app)

        response = client.get("/test/")
        assert response.status_code == 400
        assert response.content_type == "application/json"

        data = json.loads(response.data.decode("utf8"))
        assert data == {
            "message": "error",
            "test": "value",
        }

    def test_handle_api_error(self, app, client):
        api = restx.Api(app)

        @api.route("/api", endpoint="api")
        class Test(restx.Resource):
            def get(self):
                abort(404)

        response = client.get("/api")
        assert response.status_code == 404
        assert response.headers["Content-Type"] == "application/json"
        data = json.loads(response.data.decode())
        assert "message" in data

    def test_handle_non_api_error(self, app, client):
        restx.Api(app)

        response = client.get("/foo")
        assert response.status_code == 404
        assert response.headers["Content-Type"] == "text/html; charset=utf-8"

    def test_non_api_error_404_catchall(self, app, client):
        api = restx.Api(app, catch_all_404s=True)

        response = client.get("/foo")
        assert response.headers["Content-Type"] == api.default_mediatype

    def test_handle_error_signal(self, app):
        api = restx.Api(app)

        exception = BadRequest()

        recorded = []

        def record(sender, exception):
            recorded.append(exception)

        got_request_exception.connect(record, app)
        try:
            # with self.app.test_request_context("/foo"):
            api.handle_error(exception)
            assert len(recorded) == 1
            assert exception is recorded[0]
        finally:
            got_request_exception.disconnect(record, app)

    def test_handle_error_signal_does_not_call_got_request_exception(self, app):
        api = restx.Api(app)

        exception = BadRequest()

        recorded = []

        def record(sender, exception):
            recorded.append(exception)

        @api.errorhandler(BadRequest)
        def handle_bad_request(error):
            return {"message": str(error), "value": "test"}, 400

        got_request_exception.connect(record, app)
        try:
            api.handle_error(exception)
            assert len(recorded) == 0
        finally:
            got_request_exception.disconnect(record, app)

    def test_handle_error(self, app):
        api = restx.Api(app)

        response = api.handle_error(BadRequest())
        assert response.status_code == 400
        assert json.loads(response.data.decode()) == {
            "message": BadRequest.description,
        }

    def test_handle_error_does_not_duplicate_content_length(self, app):
        api = restx.Api(app)

        # with self.app.test_request_context("/foo"):
        response = api.handle_error(BadRequest())
        assert len(response.headers.getlist("Content-Length")) == 1

    def test_handle_smart_errors(self, app):
        api = restx.Api(app)
        view = restx.Resource

        api.add_resource(view, "/foo", endpoint="bor")
        api.add_resource(view, "/fee", endpoint="bir")
        api.add_resource(view, "/fii", endpoint="ber")

        with app.test_request_context("/faaaaa"):
            response = api.handle_error(NotFound())
            assert response.status_code == 404
            assert json.loads(response.data.decode()) == {
                "message": NotFound.description,
            }

        with app.test_request_context("/fOo"):
            response = api.handle_error(NotFound())
            assert response.status_code == 404
            assert "did you mean /foo ?" in response.data.decode()

        app.config["RESTX_ERROR_404_HELP"] = False

        response = api.handle_error(NotFound())
        assert response.status_code == 404
        assert json.loads(response.data.decode()) == {"message": NotFound.description}

    def test_handle_include_error_message(self, app):
        api = restx.Api(app)
        view = restx.Resource

        api.add_resource(view, "/foo", endpoint="bor")

        with app.test_request_context("/faaaaa"):
            response = api.handle_error(NotFound())
            assert "message" in json.loads(response.data.decode())

    def test_handle_not_include_error_message(self, app):
        app.config["ERROR_INCLUDE_MESSAGE"] = False

        api = restx.Api(app)
        view = restx.Resource

        api.add_resource(view, "/foo", endpoint="bor")

        with app.test_request_context("/faaaaa"):
            response = api.handle_error(NotFound())
            assert "message" not in json.loads(response.data.decode())

    def test_error_router_falls_back_to_original(self, app, mocker):
        class ProgrammingBlunder(Exception):
            pass

        blunder = ProgrammingBlunder("This exception needs to be detectable")

        def raise_blunder(arg):
            raise blunder

        api = restx.Api(app)
        app.handle_exception = mocker.Mock()
        api.handle_error = mocker.Mock(side_effect=raise_blunder)
        api._has_fr_route = mocker.Mock(return_value=True)
        exception = mocker.Mock(spec=HTTPException)

        api.error_router(app.handle_exception, exception)

        app.handle_exception.assert_called_with(blunder)

    def test_fr_405(self, app, client):
        api = restx.Api(app)

        @api.route("/ids/<int:id>", endpoint="hello")
        class HelloWorld(restx.Resource):
            def get(self):
                return {}

        response = client.post("/ids/3")
        assert response.status_code == 405
        assert response.content_type == api.default_mediatype
        # Allow can be of the form 'GET, PUT, POST'
        allow = ", ".join(set(response.headers.get_all("Allow")))
        allow = set(method.strip() for method in allow.split(","))
        assert allow == set(["HEAD", "OPTIONS", "GET"])

    @pytest.mark.options(debug=True)
    def test_exception_header_forwarded(self, app, client):
        """Ensure that HTTPException's headers are extended properly"""
        api = restx.Api(app)

        class NotModified(HTTPException):
            code = 304

            def __init__(self, etag, *args, **kwargs):
                super(NotModified, self).__init__(*args, **kwargs)
                self.etag = quote_etag(etag)

            def get_headers(self, *args, **kwargs):
                return [("ETag", self.etag)]

        custom_abort = Aborter(mapping={304: NotModified})

        @api.route("/foo")
        class Foo1(restx.Resource):
            def get(self):
                custom_abort(304, etag="myETag")

        foo = client.get("/foo")
        assert foo.get_etag() == unquote_etag(quote_etag("myETag"))

    def test_handle_server_error(self, app):
        api = restx.Api(app)

        resp = api.handle_error(Exception())
        assert resp.status_code == 500
        assert json.loads(resp.data.decode()) == {"message": "Internal Server Error"}

    def test_handle_error_with_code(self, app):
        api = restx.Api(app, serve_challenge_on_401=True)

        exception = Exception()
        exception.code = "Not an integer"
        exception.data = {"foo": "bar"}

        response = api.handle_error(exception)
        assert response.status_code == 500
        assert json.loads(response.data.decode()) == {"foo": "bar"}

    def test_handle_error_http_exception_response_code_only(self, app):
        api = restx.Api(app)
        http_exception = HTTPException(response=Response(status=401))

        response = api.handle_error(http_exception)
        assert response.status_code == 401
        assert json.loads(response.data.decode()) == {
            "message": "Unauthorized",
        }

    def test_errorhandler_swagger_doc(self, app, client):
        api = restx.Api(app)

        class CustomException(RuntimeError):
            pass

        error = api.model("Error", {"message": restx.fields.String()})

        @api.route("/test/", endpoint="test")
        class TestResource(restx.Resource):
            def get(self):
                """
                Do something

                :raises CustomException: In case of something
                """
                pass

        @api.errorhandler(CustomException)
        @api.header("Custom-Header", "Some custom header")
        @api.marshal_with(error, code=503)
        def handle_custom_exception(error):
            """Some description"""
            pass

        specs = client.get_specs()

        assert "Error" in specs["definitions"]
        assert "CustomException" in specs["responses"]

        response = specs["responses"]["CustomException"]
        assert response["description"] == "Some description"
        assert response["schema"] == {"$ref": "#/definitions/Error"}
        assert response["headers"] == {
            "Custom-Header": {"description": "Some custom header", "type": "string"}
        }

        operation = specs["paths"]["/test/"]["get"]
        assert "responses" in operation
        assert operation["responses"] == {
            "503": {"$ref": "#/responses/CustomException"}
        }

    def test_errorhandler_with_propagate_true(self, app, client):
        """Exceptions with errorhandler should not be returned to client, even
        if PROPAGATE_EXCEPTIONS is set."""
        app.config["PROPAGATE_EXCEPTIONS"] = True
        api = restx.Api(app)

        @api.route("/test/", endpoint="test")
        class TestResource(restx.Resource):
            def get(self):
                raise RuntimeError("error")

        @api.errorhandler(RuntimeError)
        def handle_custom_exception(error):
            return {"message": str(error), "test": "value"}, 400

        response = client.get("/test/")
        assert response.status_code == 400
        assert response.content_type == "application/json"

        data = json.loads(response.data.decode("utf8"))
        assert data == {
            "message": "error",
            "test": "value",
        }

    def test_namespace_errorhandler_with_propagate_true(self, app, client):
        """Exceptions with errorhandler on a namespace should not be
        returned to client, even if PROPAGATE_EXCEPTIONS is set."""
        app.config["PROPAGATE_EXCEPTIONS"] = True
        api = restx.Api(app)
        namespace = restx.Namespace("test_namespace")
        api.add_namespace(namespace)

        @namespace.route("/test/", endpoint="test")
        class TestResource(restx.Resource):
            def get(self):
                raise RuntimeError("error")

        @namespace.errorhandler(RuntimeError)
        def handle_custom_exception(error):
            return {"message": str(error), "test": "value"}, 400

        response = client.get("/test_namespace/test/")
        assert response.status_code == 400
        assert response.content_type == "application/json"

        data = json.loads(response.data.decode("utf8"))
        assert data == {
            "message": "error",
            "test": "value",
        }



import pytest

from flask_restx import utils


class MergeTestCase(object):
    def test_merge_simple_dicts_without_precedence(self):
        a = {"a": "value"}
        b = {"b": "other value"}
        assert utils.merge(a, b) == {"a": "value", "b": "other value"}

    def test_merge_simple_dicts_with_precedence(self):
        a = {"a": "value", "ab": "overwritten"}
        b = {"b": "other value", "ab": "keep"}
        assert utils.merge(a, b) == {"a": "value", "b": "other value", "ab": "keep"}

    def test_recursions(self):
        a = {
            "a": "value",
            "ab": "overwritten",
            "nested_a": {"a": "nested"},
            "nested_a_b": {"a": "a only", "ab": "overwritten"},
        }
        b = {
            "b": "other value",
            "ab": "keep",
            "nested_b": {"b": "nested"},
            "nested_a_b": {"b": "b only", "ab": "keep"},
        }
        assert utils.merge(a, b) == {
            "a": "value",
            "b": "other value",
            "ab": "keep",
            "nested_a": {"a": "nested"},
            "nested_b": {"b": "nested"},
            "nested_a_b": {"a": "a only", "b": "b only", "ab": "keep"},
        }

    def test_recursions_with_empty(self):
        a = {}
        b = {
            "b": "other value",
            "ab": "keep",
            "nested_b": {"b": "nested"},
            "nested_a_b": {"b": "b only", "ab": "keep"},
        }
        assert utils.merge(a, b) == b


class UnpackImportResponse(object):
    def test_import_werkzeug_response(self):
        assert utils.import_werkzeug_response() != None


class CamelToDashTestCase(object):
    def test_no_transform(self):
        assert utils.camel_to_dash("test") == "test"

    @pytest.mark.parametrize(
        "value,expected",
        [
            ("aValue", "a_value"),
            ("aLongValue", "a_long_value"),
            ("Upper", "upper"),
            ("UpperCase", "upper_case"),
        ],
    )
    def test_transform(self, value, expected):
        assert utils.camel_to_dash(value) == expected


class UnpackTest(object):
    def test_single_value(self):
        data, code, headers = utils.unpack("test")
        assert data == "test"
        assert code == 200
        assert headers == {}

    def test_single_value_with_default_code(self):
        data, code, headers = utils.unpack("test", 500)
        assert data == "test"
        assert code == 500
        assert headers == {}

    def test_value_code(self):
        data, code, headers = utils.unpack(("test", 201))
        assert data == "test"
        assert code == 201
        assert headers == {}

    def test_value_code_headers(self):
        data, code, headers = utils.unpack(("test", 201, {"Header": "value"}))
        assert data == "test"
        assert code == 201
        assert headers == {"Header": "value"}

    def test_value_headers_default_code(self):
        data, code, headers = utils.unpack(("test", None, {"Header": "value"}))
        assert data == "test"
        assert code == 200
        assert headers == {"Header": "value"}

    def test_too_many_values(self):
        with pytest.raises(ValueError):
            utils.unpack((None, None, None, None))


class ToViewNameTest(object):
    def test_none(self):
        with pytest.raises(AssertionError):
            _ = utils.to_view_name(None)

    def test_name(self):
        assert utils.to_view_name(self.test_none) == self.test_none.__name__


class ImportCheckViewFuncTest(object):
    def test_callable(self):
        assert callable(utils.import_check_view_func())
